#!/usr/local/bin/perl
#
# f2a.pl:    VAX Fortran-77 to Ada-95 converter toy
# Version:   1.1
# Usage:     perl f2a.pl <filename>
#            A body file <filename>.adb will be generated.
#            A spec file <filename>.ads will be generated unless
#            the Fortran file consists of just one PROGRAM.
# Author:    Oliver M. Kellogg (oliver.kellogg@vs.dasa.de)
# Copyright: (C) 1996-98, Daimler-Benz Aerospace AG (DASA), Germany
#                1999-2000, Daimler-Chrysler Aerospace AG
# 
# Ver. |  Date  | History
# -----+--------+----------------------------------------------------------
#  1.1  00/02/20  Fixed Boolean logic operators (and/or/not etc.)
#                 Fixed recognition of 'GO TO' (with a space) in one-line IF.
#                 Improved translation of array initializations, but the
#                 implied-DO in DATA is still missing.
#                 First hack at BLOCKDATA.
#  1.0  00/01/22  The subprogram parameter conversion was broken after the
#                 previous change to @arg, fixed it. Also fixed the RETURN 
#                 mechanism of FUNCTIONs by adding "_Result" to the function 
#                 return variable name.
#                 Rewrote handling of DATA; first hack at DIMENSION.
#  0.9  00/01/17  Various improvements. @arg no longer carries the keyword
#                 in $arg[0] (that's what $kw is there for.)
#  0.8  99/12/09  Improved handling of single-line IF statements.
#                 Implemented colon-separated dimension bounds in array
#                 declarations (idea by by brian.orpin@gecm.co.uk)
#  0.7  97/11/12  Very minor cosmetic changes, new email contact address
#  0.6  97/05/10  Implemented trivial cases of EQUIVALENCE.
#                 Repaired end-of-file problems, the DATA statement, and 
#                 label placement.
#                 Implemented trivial CHARACTER*n declarations.
#                 Added some more intconst-to-float conversions.
#                 Added INCLUDE-packagename and COMMON-recordname prefixing
#                 at variable references.
#                 Added specfile generation. Both spec and body are now
#                 written to files instead of the body being echoed to
#                 standard output.
#  0.5  97/05/05  Implemented COMMON and non-recursive INCLUDE
#  0.4  97/02/20  Convert integer constants to floating syntax where
#                 necessity to do so is easily identified
#  0.3  97/02/16  Eliminated usage of Interfaces.Fortran altogether;
#                 expanded package Ftn accordingly
#  0.2  97/02/11  Simplified the Ada types used
#                 (Integer instead of Fortran_Integer etc., see pkg Ftn.)
#                 Added output of outermost scope declarations (at end.)
#  0.1  96/11/26  Initial version
#
# -------------------------------------------------------------------------
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# ---------------------------------------------------------------------------
# Warnings:
# 1. The aim of this toy is NOT to replace a Fortran compiler
#    (as for example the f2c program does.) Do not expect to pass the
#    output through an Ada compiler directly without manual tuning.
# 2. Being a primitive two-pass text substitution tool, translation is
#    at best HALF automatic.
#    The tool has been tried only on some standalone numerical algorithms.
#    FORMAT directives and I/O operations are not supported at all, nor
#    are those constructs that don't map to Ada in an easy way.
#    Best results are obtained on "structured" F77 code (i.e. code that
#    minimizes use of GOTO and employs block oriented statements like
#    IF/THEN/ELSE/ENDIF and DO/ENDDO.) If you feed it spaghetti, that's
#    just what it will produce.
# ---------------------------------------------------------------------------
# Please send your suggestions/bugreports/improvements to:
# Oliver.Kellogg@vs.dasa.de

sub matching;
sub ftnfind;
sub nonarg_comma_index;
sub getline;
sub specprint;
sub bodyprint;
sub emit;
sub spaces;
sub indent;
sub dedent;
sub dent;
sub lessdent;
sub prettyprint;
sub check_begin;
sub ftnsplit;
sub get_implicit_type;
sub get_f77type_index;
sub syminsert;
sub symlookup;
sub basetype;
sub targettype;
sub output_decls;
sub add_dolabel;
sub init;
sub init_variables;
sub init_implicit;
sub do_dim_initvals;
sub arith_if_comparison;


# for ($i = 0; $i < 26; $i++) {
#     print "$i => $implic[$i]\n";
# }

$#ARGV >= 0 or die "f2a -- supply F77 input file name\n";
$input = MAIN;
open($input, $ARGV[0]) or die "Can't open input file\n";

($main_pkgname = uc $ARGV[0]) =~ s/\W.*//;
$main_pkgname =~ s/-/_/g;
$pkgname = $main_pkgname;

# All Fortran predefined types and functions are assumed to be in package Ftn,
# as are basic (unconstrained) array types like:
#    type Integer_Array is array (Integer range <>) of Integer;
#    type Real_Array is array (Integer range <>) of Real;
# and so on. All multi-dimensional arrays are assumed to be pragma Convention
# (Fortran). The standard Ada type Integer is used instead of type
# Interfaces.Fortran.Fortran_Integer in order to avoid type problems in
# array indexing. (For example, the default type of the loop variable
# in for-loops is Standard.Integer.)
# Restrictions:
# - Nothing after the END of the main program will be processed. The
#   main program should appear *after* any subroutine and function
#   declarations (i.e. short before the end of file.)
# Things missing:
# - recursive INCLUDE (top-level INCLUDE is okay)
# - implied-DO in DATA
# - BLOCKDATA is only partially implemented
# - EXTERNAL
# - lots of others that escape my mind --
#   tell me what you think are important ones.

# initializations
init;
init_implicit;

LINE:
while (($line = getline) ne "" or ! $eof) {
#  print "RAW: ",$line,"\n";

    next if ($line eq "");
    if ($line =~ /^\s*([0-9]+)\s+(.*)/) {        # target label
        my $label = $1;
        my $cmd = $2;
        check_begin;
        if ($cmd =~ /^FORMAT/i) {
            emit "-- f2a TBD: $line";
            next;
        }
        if ($#dolabels >= 0 and $dolabels[$#dolabels] == $label) {
            $pending_label = $label;
        } else {
            lessdent "<<L$label>>   null;\n";
        }
        next if ($cmd =~ /^$/ or $cmd =~ /^CONTINUE$/i);
        $line = $cmd;
    }

    if ($line =~ /^$/) {              # empty line
        next;
    } elsif ($line =~ /^[C\*](.*)/i) {    # comment
        emit("\-\-$1");
        next;
    } elsif (($exclam = ftnfind($line, '!')) >= 0) {
        $comment = substr($line, $exclam+1);
        next if ($exclam == 0);
        $line = substr($line, 0, $exclam-1);
        $line =~ s/\s+$//;
    }

    $line =~ s/^\s+//;
    next if ($line eq "");

    if ($line =~ /^INCLUDE/i) {
        # my $filename = join("", splice(@arg, 2, $#arg-2));
        my $filename = substr($line, index($line,"'") + 1, -1);
        $input = INCLUDE;
# print("include: " . $filename . "\n");
        open($input, $filename) or die "Can't open include file $filename\n";
        ($pkgname = uc $filename) =~ s/\W.*//;
        bodyprint "with $pkgname;";
        bodyprint("  -- move to top of file") unless ($first);
        bodyprint "\n";
        next;
    } elsif ($line =~ /^BLOCK\s*DATA\s*(\w+)/i) {
        $seen_any_subprogram = 1;
        $in_blockdata = 1;
        $subprogram_name = $1;
        next;
    } elsif ($line =~ /^GO\s*TO\s*(\d+)/i) {
        check_begin;
        dent "goto L$1;";
        next;
    } elsif ($line =~ /^ELSE\s*IF/i) {
        $line =~ s/ELSE\s*//i;
        $elsif = 1;                      # See IF processing.
    } elsif ($line =~ /^END\s*IF/i) {
        dedent "end if;";
        next;
    } elsif ($line =~ /^END\s*DO/i) {
        if (($stepval = pop @dosteps) != 0) {
            $varname = pop @dovars;
            dent "$varname := $varname ";
            if ($stepval < 0) {
                emit "- " . -$stepval . ";\n";
            } else {
                emit "+ " . $stepval . ";\n";
            }
        }
        dedent "end loop;";
        next;
    } elsif ($line =~ /^END\s*STRUCTURE/i) {
        $in_structure = 0;
        dedent "end record;";
        next;
    } elsif ($line =~ /^END\s*MAP/i) {
        next;
    } elsif ($line =~ /^END\s*UNION/i) {
        next;
    }

    @arg = ftnsplit $line;

#    foreach $a (@arg) {
#        print "$a|";
#    }
#    print "\n";

    # Resolve conflicts with Ada reserved words
    for ($i = 0; $i <= $#arg; $i++) {
        next if ($arg[$i] !~ /^[A-z]/);
        foreach $w (qw/
                abort access all array at begin body boolean constant
                declare delay delta digits exception exit float for generic
                in is limited long_float long_integer loop new null of 
                others out package pragma private procedure raise range
                renames reverse separate string subtype task terminate type
                until use when with /) {
            if (lc($arg[$i]) eq $w) {
                $arg[$i] = "Ftn_" . $arg[$i];
                last;
            }
        }
    }
    for ($i = 1; $i <= $#arg; $i++) {
        # Yet more Ada reserved words
        foreach $word (qw/ return case select entry accept if else elsif end/) {
            if (lc($arg[$i]) eq $word) {
                $arg[$i] = "Ftn_" . $arg[$i];
            }
        }
        if ($arg[$i] eq '*' and $arg[$i+1] eq '*') {
            # Put back together the "**" operator.
            $arg[$i] .= '*';
            splice(@arg, $i+1, 1);
        } elsif ($arg[$i] =~ /^\'.*\'$/) {
            # Convert based numbers and strings to Ada syntax
            my $contents = substr($arg[$i], 1, length($arg[$i]) - 2);
            if ($arg[$i+1] =~ /^[BOXZ]$/) {
                my $base = $arg[$i + 1];
                splice(@arg, $i + 1, 1);
                $base = (($base eq 'B') ? 2 : ($base eq 'O') ? 8 : 16);
                $arg[$i] = $base . '#' . $contents . '#';
            } else {
                $arg[$i] = '+"' . $contents . '"';
            }
        } elsif ($arg[$i] =~ /^\d/ or $arg[$i] =~ /^\.\d/) {
            # Do some float literal syntax adjustments.
            if ($arg[$i] =~ /^\./) {        # Add leading zero if need be.
                $arg[$i] = '0' . $arg[$i];
            }
            if ($arg[$i] =~ /[\.]$/) {      # Add ending zero if need be.
                $arg[$i] .= '0';
            } elsif ($arg[$i] =~ /[DE]$/i) {
                # The exponent of a floating-point literal in scientific
                # notation has been severed off, join it back.
                if ($arg[$i+1] ne '+' && $arg[$i+1] ne '-' or
                    $arg[$i+2] =~ /\D/) {
                    die("What is this: " . $arg[i] .','.
                        $arg[$i+1] .','. $arg[$i+2] . "\n");
                }
                substr($arg[$i],-1) = 'e';
                $arg[$i] .= $arg[$i+1] . $arg[$i+2];
                splice(@arg, $i+1, 2);
            } elsif ($arg[$i] =~ /[DE]/i) {
                $arg[$i] =~ s/[DE]/e/i;
            }
            if ($arg[$i] =~ /\.e/) {
                $arg[$i] =~ s/\.e/.0e/;
            }
        }
    }
    # Put unary plus/minus together with the number it belongs to.
    for ($i = 1; $i <= $#arg; $i++) {
        if ($arg[$i] =~ /^[+-]$/ and $arg[$i+1] ne '(') {
            if ($i == 1 or $arg[$i-1] =~ /^[\+\-\*\/:,(]$/) {
                $arg[$i] .= $arg[$i+1];
                splice(@arg, $i+1, 1);
            }
        }
    }
    # Compress colon (array range) usage into a single token.
    for ($i = 2; $i <= $#arg; $i++) {
        if ($arg[$i] eq ':') {
            if ($arg[$i-1] eq '(') {
                $arg[$i] = $arg[$i-2] . "'First:";
                if ($arg[$i+1] eq ')') {
                    $arg[$i] .= $arg[$i-2] . "'Last";
                } else {
                    $arg[$i] .= $arg[$i+1];
                    splice(@arg, $i + 1, 1);
                }
            } else {
                $arg[$i-1] .= ":";
                splice(@arg, $i, 1);
                if ($arg[$i] eq ')') {
                    $arg[$i-1] .= $arg[$i-3] . "'Last";
                } else {
                    $arg[$i-1] .= $arg[$i];
                    splice(@arg, $i, 1);
                }
            }
        }
    }

    # Preparation for data type processing:
    # The IMPLICIT statement takes a data type as the first argument.
    # Shifting that into $arg[0] simplifies the further processing.
    my $implicit_stmt = 0;
    if (uc($arg[0]) eq 'IMPLICIT') {
        $implicit_stmt = 1;
        shift @arg;
    }

    # Sever the keyword from @arg.
    my $kw_in_origcase = shift @arg;
    my $kw = uc $kw_in_origcase;

    # Compress builtin data type names.
    if ($kw eq 'INTEGER' or $kw eq 'REAL' or $kw eq 'DOUBLE' or
        $kw eq 'COMPLEX' or $kw eq 'CHARACTER' or $kw eq 'LOGICAL') {
        $decldim = "";
        if ($arg[0] eq '*') {
            if ($arg[1] eq '(') {
                if ($w ne 'CHARACTER') {
                    print "cannot handle declaration at $line\n";
                }
                $decldim = $arg[2];
                splice(@arg, 0, 4);
            } else {
                $kw .= $arg[1];
                splice(@arg, 0, 2);
            }
        }
        if ($arg[0] eq '(' && !$implicit_stmt) {
            if ($decldim ne "") {
                print "declaration too complex at $line\n";
            }
            $decldim = $arg[1];
            splice(@arg, 0, 3);
        }
    }

    ###################### MAIN ANALYSIS CODE ########################

    if ($implicit_stmt) {
        init_implicit;
        if ($kw eq 'NONE') {
            $implicit_none = 1;
            next;
        }
        $implicit_none = 0;
        my $type = get_f77type_index($kw);
        # print "implicit typeindex = $type\n";
        if ($type == $NONE) {
            print "Don't know datatype at: $line\n";
            next;
        }
        if ($arg[0] eq '(') {
            shift @arg;
            pop @arg;
        }
        my $i;
        for ($i = 0; $i <= $#arg; $i++) {
            next if ($arg[$i] eq ',');
            my $ndx = ord(uc $arg[$i]) - 65;
            if ($ndx < 0 or $ndx > 25) {
                die "startletter problem in line: $line\n";
            }
            my $endx = $ndx;
            if ($arg[$i+1] eq '-') {
                $i += 2;
                $endx = ord(uc $arg[$i]) - 65;
            }
            my $j;
            for ($j = $ndx; $j <= $endx; $j++) {
                $implic[$j] = $type;
            }
# print("implic = ", join(' ',@implic));
        }
        next;
    }

    if ($kw eq 'PROGRAM' or $kw eq 'SUBROUTINE' or $kw eq 'FUNCTION') {
	if ($kw eq 'PROGRAM') {
            $seen_mainprog = 1;
        } else {
            if ($kw eq 'FUNCTION') {
                $functiontype = get_implicit_type($arg[0]);
                $functions{uc $arg[0]} = $functiontype;
            }
            $seen_any_subprogram = 1;
        }
        $subprogram_name = shift @arg;
        next if (scalar(@arg) < 3);
        if (shift(@arg) ne '(') {
            print "expecting ( at line $line\n";
            next;
        } elsif (pop(@arg) ne ')') {
            print "expecting ) at line $line\n";
            next;
        }
        my $i = 0;
        while (@arg) {
            $param[$i++] = syminsert(shift(@arg), $NONE, "", 0, "", 0, 0);
            if (scalar(@arg) && shift(@arg) ne ',') {
                print "expecting , at line $line\n";
                last;
            }
        }
        next;
    }

    if ($first and $pkgname eq $main_pkgname) {
        # print "\npackage body $pkgname is    -- at $line\n\n";
        bodyprint "\npackage body $pkgname is\n\n";
        $first = 0;
    }

    $type = get_f77type_index($kw);
    if ($type != $NONE) {    # data (or function) declaration
        my $i;
        # See if it's a function
        if ($arg[0] =~ /^FUNCTION$/i) {
            my $i;
            my $j = 0;
            init_variables;
            $functiontype = $type;
            $seen_any_subprogram = 1;
            $subprogram_name = $arg[1];
            $functions{uc $subprogram_name} = $functiontype;
            next if ($#arg < 2);
            for ($i = 3; $i < $#arg; $i += 2) {
                $param[$j++] = syminsert($arg[$i], $NONE, "", 0, "", 0, 0);
            }
            next;
        }

        # Not a function, so data declaration
        for ($i = 0; $i <= $#arg; $i++) {
            my $name = $arg[$i];
            my $dimcnt = 0;
            my @dim = ();
            next if ($arg[$i] eq ',');
            if ($name eq '(') {    # hack for missing feature...
                my $parenlvl = 1;
                my $j;
                for ($j = $i + 1; $j < $#arg; $j++) {
                    if ($arg[$j] eq '(') {
                        $parenlvl++;
                    } elsif ($arg[$j] eq ')') {
                        $parenlvl--;
                        last if (!$parenlvl);
                    }
                }
                $name = join('', splice(@arg, $i, $j - $i + 1));
            }
            if ($decldim && $type != $STRING) {
                $dim[$dimcnt++] = $decldim;
            }
            if ($i < $#arg and $arg[$i+1] eq '(') {    # dimension
                for ($i = $i + 2; $arg[$i] ne ')'; $i++) {
                    next if ($arg[$i] eq ',');
                    if ($arg[$i] !~ /:/) {
                        $arg[$i] = "1:" . $arg[$i];
                    }
                    $dim[$dimcnt++] = $arg[$i];
                }
            }
            my $initstr = "";
            if ($arg[$#arg] eq '/') {                  # initial value
                while ($arg[$i] ne '/') {
                    $i++;
                }
                $i++;
                $initstr = join('', splice(@arg, $i, $#arg - $i));
            }
            if ($in_structure) {
                my $ttype = targettype($type, \@dim);
                dent "$name : $ttype;";
            } elsif ($kw eq 'DATA') {     # Modify the init string
                syminsert($name, $NONE, $initstr, 0, "", 1, 0);
            } else {
                my $is_assigned = ($initstr ne "");
                syminsert($name, $type, $initstr, 0, "", $is_assigned,
                           $dimcnt, \@dim);
            }
        }
        next;
    }

    if ($kw eq 'DATA') {    # implied-DO is not yet implemented
        my @names = ();
        while (@arg) {
            my $name = shift @arg;
            if ($name eq '(') {    # oops, To Be Done...
                print "implied-DO in DATA statement is not yet implemented\n";
                dent "-- f2a message: please correct the following line(s) -\n";
                dent("$kw (" . prettyprint(\@arg) . "\n");
                dent "-- f2a message end -----------------------------------";
                last;
            }
            push @names, $name;
            my $nxt = shift @arg;
            next if ($nxt eq ',');
            if ($nxt ne '/') {
                print "expecting either , or / in line $line\n";
                last;
            }
            foreach $name (@names) {
                my $initstr = "";
                my $symindex = symlookup($name);
                if ($symindex >= 0 && @{$symdim[$symindex]}) {
                    do_dim_initvals($symdim[$symindex], \@arg, \$initstr);
                }
                if (! $initstr) {
                    while (scalar(@arg) && $arg[0] ne '/' && $arg[0] ne ',') {
                        $initstr .= shift @arg;
                    }
                }
                if ($in_blockdata) {
                    check_begin;
                    dent;
                    if ($sympkg[$symindex] ne $main_pkgname) {
                        emit($sympkg[$symindex] . '.');
                    }
                    if ($symcommon[$symindex] ne "") {
                        emit($symcommon[$symindex] . '.');
                    }
                    emit($name . " := $initstr;\n");
                } else {
                    syminsert($name, $NONE, $initstr, 0, "", 1, 0);
                }
                shift(@arg) if ($arg[0] eq '/');
                shift(@arg) if ($arg[0] eq ',');
            }
            @names = ();
        }
        next;
    }

    if ($kw eq 'DIMENSION') {
        loop_over_all_declared_variables:
        while (@arg) {
            my $name = shift @arg;
            my $dimcnt = 0;
            my @dim = ();
            if (shift(@arg) ne '(') {
                print "expecting ( in line $line\n";
                last;
            }
            my $terminator;
            loop_over_dimensions_for_one_variable:
            do {
                my $d = shift @arg;
                if ($d !~ /:/) {
                    $d = "1:" . $d;
                }
                $dim[$dimcnt++] = $d;
                $terminator = shift @arg;
            } while ($terminator eq ',');
            if ($terminator ne ')') {
                print "expecting ) in line $line\n";
                last;
            }
            syminsert($name, $NONE, "", 0, "", 0, $dimcnt, \@dim);
            last if (! @arg);
            if (shift(@arg) ne ',') {
                print "expecting , in line $line\n";
                last;
            }
        }
        next;
    }

    if ($kw eq 'COMMON') {
        if (shift(@arg) ne '/') {
            print "expecting / at line $line\n";
            next;
        }
        my $comname = shift @arg;
        if (shift(@arg) ne '/') {
            print "expecting second / at line $line\n";
            next;
        }
        my $i;
        my @indexes = ();
        my $idxcount = 0;
        while (@arg) {
            my $name = shift @arg;
            my $dimcnt = 0;
            my @dim = ();
            if (@arg) {
                my $nxt = shift @arg;
                if ($nxt eq '(') {    # dimension
                    while (@arg && $arg[0] ne ')') {
                        my $item = shift @arg;
                        next if ($item eq ',');
                        $dim[$dimcnt++] = $item;
                    }
                    shift @arg;   # discard )
                    if (@arg) {
                        if (shift(@arg) ne ',') {
                            print "expecting , at line $line\n";
                        }
                    }
                } elsif ($nxt ne ',') {
                    print "expecting , at line $line\n";
                }
            }
            $indexes[$idxcount++] = syminsert($name, $NONE, "", 0, $comname, 0,
                                               $dimcnt, \@dim);
        }
        $commons{$comname} = \@indexes;
        $commain{$comname} = ($pkgname eq $main_pkgname);
        next;
    }

    if ($kw eq 'STRUCTURE') {
        indent "type " . $arg[1] . " is record";
        $in_structure = 1;
        next;
    }

    if ($kw eq 'RECORD') {           # not quite ready yet!
        my $type = $arg[1];
        my $i;
        for ($i = 3; $i <= $#arg; $i++) {
            my $name = $arg[$i];
            my $dimcnt = 0;
            my @dim = ();
            next if ($arg[$i] eq ',');
            if ($i < $#arg and $arg[$i+1] eq '(') {    # dimension
                for ($i = $i + 2; $arg[$i] ne ')'; $i++) {
                    next if ($arg[$i] eq ',');
                    $dim[$dimcnt++] = $arg[$i];
                }
            }
            my $initstr = "";
            if ($i < $#arg and $arg[$i+1] eq '/') {    # initial value
                for ($i = $i + 2; $arg[$i] ne '/'; $i++) {
                    $initstr .= $arg[$i];
                }
            }
            dent "$name : ";
            if ($dimcnt > 0) {
                emit "array (";
                my $d;
                for ($d = 0; $d < $dimcnt; $d++) {
                    emit(',') if ($d > 0);
                    emit "1.." . $dim[$d];
                }
                emit ") of ";
            }
            emit $type;
            if ($initstr ne "") {
                emit " := ";
                emit '(';
                emit $initstr;
                emit ')';
            }
            emit ";\n";
        }
        next;
    }

    if ($kw eq 'PARAMETER') {
        my $i;
        for ($i = 1; $i < $#arg; $i++) {
             my $name = $arg[$i];
             my $initval = "";
             for ($i = $i + 2; $i < $#arg; $i++) {
                 last if ($arg[$i] eq ',');
                 $initval .= $arg[$i];
             }
             syminsert($name, $NONE, $initval, 1, "", 1, 0);
        }
        next;
    }

    if ($kw eq 'EQUIVALENCE') {
        my $original = $arg[1];
        my $newname = $arg[-2];
        if ($arg[2] eq '(') {
            if ($arg[-2] eq ')') {
                bodyprint "-- f2a TBD: $line\n";
                next;
            }
            $original = $arg[1];
            my $i = 2;
            do {
                $original .= $arg[$i];
            } while ($arg[$i++] ne ')');
        } elsif ($arg[4] eq '(') {
            $original = $arg[3];
            my $i = 4;
            do {
                $original .= $arg[$i];
            } while ($arg[$i++] ne ')');
            $newname = $arg[1];
        }
        my $symindex = symlookup($newname);
        if ($symindex >= 0 && @{$symdim[$symindex]}) {
            bodyprint "-- f2a TBD: $line\n";
            next;
        }
        syminsert($newname, $NONE, ('=' . $original), 0, "", 0, 0);
        # The '=' at the beginning of the initval is the special sign
        # for EQUIVALENCE. This is decoded in sub output_decl.
        next;
    }

    if ($pkgname ne $main_pkgname) {
        print "-- in included $pkgname: discarding  $line\n";
        next;
    }

    if ($kw eq 'END') {
        my $i;
        my $symindex;
        my $name;
        $direct_output = $TO_BODY;      # emit to body
        if ($seen_any_subprogram) {
           $direct_output |= $TO_SPEC;  # emit also to spec
        }
        $indentlevel = 1;
        if ($functiontype != $NONE) {
            dent "function ";
        } else {
            dent "procedure ";
        }
        emit "$subprogram_name ";
        if ($#param >= 0) {
            emit "(\n";
            $indentlevel += 6;
            for ($i = 0; $i <= $#param; $i++) {
                $symindex = $param[$i];
                $name = $symname[$symindex];
                dent "$name : ";
                if ($symwritten[$symindex]) {
                    emit "in out ";
                }
                # if ($symdim[$symindex] && @$symdim[$symindex]) {
                #     specprint("Ftn.");
                # }
                emit basetype($symtype[$symindex], $symdim[$symindex]);
                if ($i < $#param) {
                   emit ";\n";
                } else {
                   emit ") ";
                }
            }
            $indentlevel -= 6;
        }
        if ($functiontype != $NONE) {
            emit("return " . $adatype[$functiontype] . " ");
        }
        specprint(";\n\n") if ($direct_output & $TO_SPEC);
        $direct_output = $TO_BODY;  # emit only to body
        emit "is\n";
        indent;
        # output local declarations
        output_decls;
        bodyprint $outbuf;
        if ($functiontype != $NONE) {
            # Make sure we leave with a RETURN statement.
            # It is needed here due to Fortran return conventions.
            # For beauty's sake, we output it only if the last statement
            # wasn't a RETURN statement anyway (otherwise, we'd have two
            # RETURNs in a row.)
            my $last = substr($outbuf, -80);
            $last =~ s/.*\n(.*)\n$/$1/;
            if ($last !~ / return /) {
                dent "return $subprogram_name;\n";
            }
        }
        dedent "end $subprogram_name;\n\n";
        init_variables;
        next;
    }

HALF_COOKED:

    if ($kw eq 'GOTO') {   # this happens when we got here via (perl) goto
        dent("goto L" . $arg[0] . ";");
        next;
    }

    if ($kw eq 'CALL') {
        check_begin;
        dent shift(@arg);
        if (@arg) {
            emit(" " . prettyprint(\@arg));
        }
        emit ';';
        next;
    }

    if ($kw eq 'RETURN') {
        check_begin;
        if ($functiontype != $NONE) {
            dent "return ${subprogram_name}_Result;";
            next;
        }
        if ($line_pending) {
            next if ($linebuf =~ /END$/i);   # avoid redundant statement
        }
        dent "return;";
        next;
    }

    if ($kw eq 'DO') {
        check_begin;
        if ($arg[0] =~ /^\d/) {    # we have a do-label
            push @dolabels, shift @arg;
            if ($arg[0] eq ',') {
                shift @arg;
            }
        }
        if (uc($arg[0]) eq 'WHILE') {
            shift @arg;  # discard WHILE
            shift @arg;  # discard (
            pop @arg;    # discard )
            indent("while " . prettyprint(\@arg) . " loop");
            next;
        }
        my $varname = shift @arg;
        if (shift(@arg) ne '=') {
            print "Huh? '=' expected in line $line\n";
        }
        my $commapos;
        if (($commapos = nonarg_comma_index(\@arg)) < 0) {
            print "expecting comma between start and end value at $line\n";
            next;
        }
        my $startval = join("", splice(@arg, 0, $commapos));
        shift @arg;     # discard the comma
        my $in_reverse = 0;
        my $endval;
        my @stepval = ();
        if (($commapos = nonarg_comma_index(\@arg)) >= 0) {
            my $forget_stepval = 0;
            $endval = join("", splice(@arg, 0, $commapos));
            shift @arg;     # discard the comma
            @stepval = @arg;
            if (scalar(@stepval) == 1) {
                if ($stepval[0] == 1) {
                    $forget_stepval = 1;
                } elsif ($stepval[0] == -1) {
                    $forget_stepval = 1;
                    $in_reverse = 1;
                }
            }
            unless ($forget_stepval) {
                push @dosteps, join("", @stepval);
                push @dovars, $varname;
                dent("$varname := $startval;\n");
                my $comparison = "<=";
                if ($stepval[0] =~ /^\-/) {
                    $comparison = ">=";
                }
                indent("while $varname $comparison $endval loop");
                next;
            }
        } else {
            $endval = join("", @arg);
        }
        # We'll use a for-loop if no step value is given, or if the step
        # value is 1 or -1.  That's not always correct, but a best effort.
        push @dosteps, 0;
        if ($in_reverse) {
            indent "for $varname in reverse $endval .. $startval loop";
        } else {
            indent "for $varname in $startval .. $endval loop";
        }
        next;
    }

    # The famous case of a variable named "IF"
    if ($kw eq 'IF' && $arg[0] eq '=') {
        $kw_in_origcase = "Ftn_$kw_in_origcase";
        $kw = uc $kw_in_origcase;
    }

    if ($kw eq 'IF') {
        check_begin;
        # discard keyword
        my $parenlvl = 0;
        my @cond = ();
        # discard opening parenthesis
        if (length($arg[0]) > 1) {
            $arg[0] = substr($arg[0], 1);
        } else {
            shift @arg;
        }
        # Move the condition from @arg to @cond
        while (@arg) {
            my $item = $arg[0];
            if ($item eq ')') {
                if ($parenlvl == 0) {
                    shift @arg;
                    last;
                }
                $parenlvl--;
            } elsif ($item eq '(') {
                $parenlvl++;
            }
            push @cond, shift @arg;
        }
        my $cond = prettyprint \@cond;
        if ($elsif) {
            lessdent "elsif $cond ";
            $elsif = 0;
        } else {
            indent "if $cond ";
        }
        if ($arg[0] =~ /^THEN$/i) {    # multi-line IF-THEN/ENDIF
            if ($#arg > 0) {
                print("Oops, unexpected stuff after THEN at: "
                      . $line . "\n");
            }
            emit "then";
            next;
        }
        # one-line IF
        if ($arg[0] =~ /^\d/) {               # arithmetic IF
            my $buf = getline;
            if ($buf !~ /^\s*(\d+)\s+(.*)$/) {
                emit "-- Arithmetic-IF not processed here: $line\n";
                dedent;
                $line = $buf;
                redo LINE;
            }
            my $next_label = $1;
            $buf = $2;
            # @labels = split(/,/, join("", @arg));
            @labels = ($arg[0], $arg[2], $arg[4]);
            my $necessary_label = arith_if_comparison(@labels,$next_label);
            if ($necessary_label == 0) {
                emit "-- Arithmetic-IF not processed here: $line\n";
                dedent;
                $line = $buf;
                redo LINE;
            }
            emit "$arith_if_operator 0 then\n";
            dent "goto L$necessary_label;\n";
            dedent "end if;\n";
            $line = $buf;
            redo LINE;
        }
        emit "then\n";
        $kw = shift @arg;
        if (uc($kw) eq 'GO' && uc($arg[0]) eq 'TO') {
            $kw = 'GOTO';
            shift @arg;
        }
        $kw =~ s/^Ftn_//;   # undo reserved-word conflict resolution
        $line = "$kw " . prettyprint(\@arg);
        $kw_in_origcase = $kw;
        $kw = uc $kw;
        $pending_endif = 1;
        # Whoa, here comes the dreaded GOTO
        goto HALF_COOKED;
    }

    if ($kw eq 'ELSE') {
        lessdent "else";
        next;
    }

    my @tbd = qw/ WRITE READ ACCEPT OPEN CLOSE INQUIRE EXTERNAL SAVE ENTRY /;
    my $done = 0;
    foreach $resword (@tbd) {
        if ($kw eq $resword) {
            $done = 1;
            dent "-- f2a TBD: $line";
            last;
        }
    }
    next if $done;

    if ($line =~ '=') {                        # assignment
        check_begin;
        dent;
        # Add LHS symbol to symbol table if it's not already there.
        # This is needed when we have IMPLICIT variables, and also
        # in order to determine the mode (in / in out) of parameters.
        if ($kw eq uc($subprogram_name)) {
            $kw_in_origcase .= "_Result";
        }
        my $nudename = $kw_in_origcase;
        $nudename =~ s/^(\w+).*/$1/;
        my $symindex = syminsert($nudename, $NONE, "", 0, "", 1, 0);
        if ($sympkg[$symindex] ne $main_pkgname) {
            emit($sympkg[$symindex] . '.');
        }
        if ($symcommon[$symindex] ne "") {
            emit($symcommon[$symindex] . '.');
        }
        emit $kw_in_origcase;
        while ($arg[0] ne '=') {
            emit(shift @arg);
        }
        shift @arg;
        emit " := ";
        emit prettyprint(\@arg, ($symtype[$symindex] >= $REAL));
        emit ';';
        next;
    }

    # We shouldn't be getting here
    print "$kw ";
    for ($i = 0; $i <= $#arg; $i++) {
        print $arg[$i],"|";
    }
    print "   -- f2a UFO\n";

} continue {
    if ($comment ne "") {
        emit "   --$comment";
        $comment = "";
    }
    emit "\n";
    if ($pending_endif) {
        dedent "end if;\n";
        $pending_endif = 0;
    }
    if ($pending_label) {
        while ($dolabels[$#dolabels] == $pending_label) {
            my $stepval;
            pop @dolabels;
            if (($stepval = pop @dosteps) != 0) {
                my $varname = pop @dovars;
                dent "$varname := $varname ";
                if ($stepval < 0) {
                    emit "- " . -$stepval . ";\n";
                } else {
                    emit "+ " . $stepval . ";\n";
                }
            }
            dedent "end loop;\n";
        }
        $pending_label = 0;
    }
    if ($eof) {
        last if ($pkgname eq $main_pkgname and $subprogram_name eq "");
        if ($input == INCLUDE) {
            $eof = 0;
            close $input;
            $input = MAIN;
            $pkgname = $main_pkgname;
        }
    }
}

if ($#symname >= 0) {   # means we haven't had real body output
    $direct_output = $TO_SPEC;  # emit to spec
    output_decls;
}
# print("OUTBUF:\n" . $outbuf . "\nEND-OUTBUF\n") if (length($outbuf) > 0);
if ($opened_specfile) {
    print SPEC "\nend $pkgname;\n";
    close SPEC;
}
if ($opened_bodyfile) {
    print BODY "\nend $pkgname;\n";
    close BODY;
}
close MAIN;

# ------------------------ END OF MAIN PROGRAM -------------------------

#                             SUBROUTINES

sub ftnsplit {
    my $str = shift;
    my $in_string = 0;
    my $in_lit = 0;
    my $in_space = 0;
    my $i;
    my @out = ();
    my $ondx = -1;
    for ($i = 0; $i < length($str); $i++) {
        my $ch = substr($str, $i, 1);
        if ($in_string) {
            if ($ch eq "'") {
                if (substr($str, $i+1, 1) eq "'") {
                    $i++;             # Fortran embedded '
                } else {
                    $in_string = 0;
                }
            } elsif ($ch eq '"') {
                $out[$ondx] .= '"';   # Ada escape for embedded "
            }
            $out[$ondx] .= $ch;
        } elsif ($ch eq "'") {
            $in_string = 1;
            $out[++$ondx] = $ch;
        } elsif ($ch =~ /[a-z_0-9\.]/i) {
            my $oldi = $i;
            if ($ch eq '.') {
                # Check for comparisons (.EQ. etc.)
                for (substr($str, $i+1)) {
                    /^( *EQ *\.)/i    and $i += length($1),
                                          $out[++$ondx] = "=",     last;
                    /^( *NE *\.)/i    and $i += length($1),
                                          $out[++$ondx] = "/=",    last;
                    /^( *GT *\.)/i    and $i += length($1),
                                          $out[++$ondx] = ">",     last;
                    /^( *LT *\.)/i    and $i += length($1),
                                          $out[++$ondx] = "<",     last;
                    /^( *GE *\.)/i    and $i += length($1),
                                          $out[++$ondx] = ">=",    last;
                    /^( *LE *\.)/i    and $i += length($1),
                                          $out[++$ondx] = "<=",    last;
                    /^( *OR *\.)/i    and $i += length($1),
                                          $out[++$ondx] = "or",    last;
                    /^( *AND *\.)/i   and $i += length($1),
                                          $out[++$ondx] = "and",   last;
                    /^( *NOT *\.)/i   and $i += length($1),
                                          $out[++$ondx] = "not",   last;
                    /^( *EQV *\.)/i   and $i += length($1),
                                          $out[++$ondx] = "=",     last;
                    /^( *NEQV *\.)/i  and $i += length($1),
                                          $out[++$ondx] = "/=",    last;
                    /^( *TRUE *\.)/i  and $i += length($1),
                                          $out[++$ondx] = "true",  last;
                    /^( *FALSE *\.)/i and $i += length($1),
                                          $out[++$ondx] = "false", last;
                    last;
                }
            }
            if ($i != $oldi) {
                $in_lit = 0;
            } else {
                if (! $in_lit) {
                    $in_lit = 1;
                    $ondx++;
                }
                $out[$ondx] .= $ch;
            }
        } elsif ($in_lit) {
            $in_lit = 0;
            if ($ch !~ /\s/) {
                $out[++$ondx] = $ch;
            }
        } elsif ($ch !~ /\s/) {
            $out[++$ondx] = $ch;
        }
    }
    # For simplification of further processing,
    # turn two-word directives into a single word.
    # E.g. 'DOUBLE PRECISION' => 'DOUBLE'
    for ($i = 0; $i < $#out - 1; $i++) {
        if (uc($out[$i]) eq 'DOUBLE' && uc($out[$i+1]) eq 'PRECISION') {
            splice @out, $i + 1, 1;
            last;
        }
    }
    @out;
}


# Find position of corresponding closing parenthesis of the
# first parenthesis found in string

sub matching {
    my $str = shift;
    my $paren_count = 1;
    my $pos = index($str,'(');
    my $in_string = 0;
    my $i;
    
    if ($pos < 0) {
        return -1;
    }
    for ($i = $pos+1; $i < length($str); $i++) {
        my $c = substr($str,$i,1);
        if ($c eq "'") {
            $in_string = ! $in_string;
            next;
        }
        next if $in_string;
        if ($c eq "(") {
            $paren_count++;
        } elsif ($c eq ")") {
            if ($paren_count <= 1) {
                return $i;
            }
            $paren_count--;
        }
    }
    -1;
}


# Find first occurrence of character in string.  This is like Perl's
# index(), but excludes Fortran strings from the search.

sub ftnfind {
    my $line = shift;
    my $seekchar = shift;
    my $in_string = 0;
    my $i;

    if ($line eq "") {
        return -1;
    }
    for ($i = 0; $i < length($line); $i++) {
        my $c = substr($line, $i, 1);
        if ($c eq '\'') {
            $in_string = ! $in_string;
        }
        next if $in_string;
        if ($c eq $seekchar) {
            return $i;
        }
    }
    -1;
}


sub nonarg_comma_index {
    my $argref = shift;
    my @arg = @{$argref};
    my $parenlvl = 0;
    my $i;
    for ($i = 0; $i <= $#arg; $i++) {
        my $a = $arg[$i];
        if ($a eq '(') {
            $parenlvl++;
        } elsif ($a eq ')') {
            $parenlvl--;
        } elsif ($a eq ',' && $parenlvl == 0) {
            return $i;
        }
    }
    -1;
}


# input line retrieval subroutines:
#  getl/ungetl (low-level)
#  getline (with continuation-line preprocessing)

$had_eof = 0;

sub ungetl {
    $linebuf = shift;
    $line_pending = 1;
    if ($had_eof) {
        $had_eof = 0;
        $eof = 1;
    } elsif ($eof and $linebuf ne "") {
        $had_eof = 1;
        $eof = 0;
    }
}

sub getl {
    my $l = "";
    if ($line_pending) {
        $line_pending = 0;
        # print("getl: line_pending, returning $linebuf\n");
        return $linebuf;
    } elsif ($eof) {
        # print("getl: \$eof, returning $l\n");
        return $l;
    } else {
        $l = <$input>;
        chop $l;
        $l =~ s/^\s+$//;
        if (eof $input) {
        # print("getl: eof detected, ");
            $eof = 1;
        }
        # print("getl: returning $l\n");
        return $l;
    }
}

sub getline {
    my $nxt;
    my $line = getl;

    while (1) {
        $nxt = getl;
        if ($nxt =~ /^     \S(.*)/) {
            # print "getline: appending $1\n";
            my $trimmed = $1;
            $trimmed =~ s/^\s+//;
            $line .= ' ' . $trimmed;      # continuation line
        } else {
            # print "getline: ungetting $nxt\n";
            ungetl $nxt;
            last;
        }
    }
    $line =~ s/\s+$//;
    # print "getline: returning $line\n";
    $line;
}

# output routines

sub bodyprint {
    my $str = shift;

    if (! $opened_bodyfile) {
        my $bodyfilename = lc($main_pkgname) . '.adb';
        open(BODY, ">$bodyfilename") or die "Can't create body file\n";
        print BODY "with Ftn;\n";
        print BODY "use  Ftn;\n\n";
        $opened_bodyfile = 1;
    }
    print BODY $str;
}

sub specprint {
    my $str = shift;
    if (! $opened_specfile) {
        my $specfilename = lc($main_pkgname) . '.ads';
        open(SPEC, ">$specfilename") or die "Can't create spec file\n";
        print SPEC "with Ftn;  use Ftn;\n\n";
        print SPEC "package $main_pkgname is\n\n";
        $opened_specfile = 1;
    }
    print SPEC $str;
}

sub emit {
    my $str = shift;
    if ($pkgname eq $main_pkgname) {
        if ($direct_output) {
            specprint($str) if ($direct_output & $TO_SPEC);
            bodyprint($str) if ($direct_output & $TO_BODY);
        } else {
            unless ($str eq "\n" && substr($outbuf, -1) eq "\n") {
                $outbuf .= $str;
            }
        }
    }
}

sub indent {
    my $optional_string = shift;
    if ($optional_string) {
        dent $optional_string;
    }
    $indentlevel++;
}

sub dedent {
    my $optional_string = shift;
    if ($indentlevel > 0) {
        $indentlevel--;
    }
    if ($optional_string) {
        dent $optional_string;
    }
}

sub spaces {
    ' ' x (3 * $indentlevel);
}

sub dent {
    my $str = shift;
    emit spaces;
    emit $str if ($str);
}

sub lessdent {
    dedent shift;
    indent;
}

sub make_floating {
    my $arg = shift;
    if ($arg =~ /^[0-9]/) {
        my $pos = index($arg, '.');
        if ($pos < 0) {
            $pos = index(uc $arg, 'E');
            if ($pos >= 0) {
                substr($arg, $pos, 0) = '.0';
            } else {
                $arg .= '.0';
            }
        }
    }
    $arg;
}

sub prettyprint {      # Nicely format a list containing expression symbols.
    my $argref = shift;
    my $is_float = 0;
    if (@_) {
        $is_float = shift;
    }
    my @arg = @$argref;
    if ($is_float && $#arg == 0) {
        # Conversion of integer constant to floating syntax is done only
        # in the most trivial case where just a single integer constant
        # is assigned to a variable of floating point type.
        $arg[0] = make_floating($arg[0]);
    }
    my $i;
    my $linecnt = 1;
    my $result = "";
    for ($i = 0; $i <= $#arg; $i++) {
        if ($arg[$i] =~ /^[A-z]/) {
            my $symindex = symlookup($arg[$i]);
            if ($symindex >= 0) {
                if ($symtype[$symindex] >= $REAL && $i < $#arg-1) {
                    if ($arg[$i+1] =~ /^[\+\-\*\/=<>]$/) {
                        $arg[$i+2] = make_floating($arg[$i+2]);
                    }
                }
                if ($sympkg[$symindex] ne $main_pkgname) {
                    $result .= $sympkg[$symindex] . '.';
                }
                if ($symcommon[$symindex] ne "") {
                    $result .= $symcommon[$symindex] . '.';
                }
            } elsif (uc($arg[$i]) eq "MOD" && $arg[$i + 1] eq '(') {
                $arg[$i] = '"' . $arg[$i] . '"';
            } elsif (exists $functions{uc($arg[$i])} and $arg[$i+1] ne '(') {
                $arg[$i] .= "_Result";
            }
        } elsif ($arg[$i] eq '(' && $arg[$i+1] eq ')') {
            splice(@arg, $i, 2);
            if (substr($result, -1) ne ' ') {
                $result .= ' ';
            }
            $i--;   # counteract loop increment
            next;   # reevaluate
        } elsif ($arg[$i] eq '/' && $arg[$i+1] eq '/') {
            $arg[$i] = '&';
            splice(@arg, $i + 1, 1);
            # Make life easier for the Ada compiler
            if (substr($arg[$i-1], -1) eq '"') {
                $arg[$i-1] = '"+"(' . substr($arg[$i-1], 1) . ')';
            }
            if (substr($arg[$i+1], -1) eq '"') {
                $arg[$i+1] = '"+"(' . substr($arg[$i+1], 1) . ')';
            }
        }
        $result .= $arg[$i];
        if ($arg[$i] eq '-') {
            next if ($i == 0 or $arg[$i-1] eq '(');
        }
        next if ($arg[$i] eq '**');
        last if ($i == $#arg);
        next if ($arg[$i+1] eq '**' || $arg[$i+1] eq ',');
        if (length($result) + length($arg[$i+1]) > $linecnt * 70) {
            $result .= "\n   " . spaces;
            $linecnt++;
        }
        if ($arg[$i+1] !~ /[(),]/) {
            $result .= ' ' if ($arg[$i] ne '(');
        }
    }
    $result;
}


# symbol management subroutines: syminsert, symlookup

sub syminsert {
    my $name = shift;
    my $type = shift;
    my $initstr = shift;
    my $is_constant = shift;
    my $comname = shift;
    my $is_assigned = shift;
    my $dimcount = shift;
    my $index = symlookup($name);
    my @dimcopy = ();
    my @initcopy = ();
    if ($dimcount > 0) {
        my $dimref = shift;
        @dimcopy = @$dimref;
    }
    if ($index >= 0) {
        if ($type != $NONE) {
            $symtype[$index] = $type;
        }
        if ($initstr ne "") {
            $syminit[$index] = $initstr;
        }
        if ($is_constant) {
            $symisconst[$index] = 1;
        }
        if ($comname ne "") {
            $symcommon[$index] = $comname;
        }
        if ($is_assigned) {
            $symwritten[$index] = 1;
        }
        if ($dimcount > 0) {
            $symdim[$index] = \@dimcopy;
        }
        return $index;
    }
    $symname[$n_symbols] = $name;
    if ($type == $NONE) {
        $symtype[$n_symbols] = get_implicit_type($name);
    } else {
        $symtype[$n_symbols] = $type;
    }
    $syminit[$n_symbols] = $initstr;
    $symisconst[$n_symbols] = $is_constant;
    $symcommon[$n_symbols] = $comname;
    $symwritten[$n_symbols] = $is_assigned;
    $symdim[$n_symbols] = \@dimcopy;
    $sympkg[$n_symbols] = $pkgname;
    ++$n_symbols - 1;
}

sub symlookup {
    my $name = uc shift;
    my $i;
    for ($i = 0; $i <= $#symname; $i++) {
        return($i) if ($name eq uc $symname[$i]);
    }
    -1;
}

sub output_decl {
    my $symindex = shift;
    my $initval = $syminit[$symindex];

    dent($symname[$symindex] . " : ");
    if ($symisconst[$symindex]) {
        emit "constant ";
    }
    emit targettype($symtype[$symindex], $symdim[$symindex]);
    if ($initval ne "") {
        if (substr($initval, 0, 1) eq '=') {
            emit " renames ";
            my $origname = substr($initval, 1);
            my $nudename = $origname;
            $nudename =~ s/\(.*//;
            my $origindex = symlookup($nudename);
            $origindex >= 0 or die "output_decl: EQUIVALENCE related error";
            if ($symcommon[$origindex] and $symcommon[$origindex] ne "") {
                emit($symcommon[$origindex] . '.') ;
            }
            emit($origname . ";  -- may need moving");
        } else {
            emit " := ";
            if (length($initval) > 60) {
                my $first = 1;
                while (1) {
                    my $lastcomma = rindex(substr($initval,0,60), ',');
                    if ($lastcomma >= 0) {
                        emit("\t") unless ($first);
                        emit(substr($initval,0,$lastcomma+1) . "\n");
                        $initval = substr($initval,$lastcomma+1);
                        $first = 0;
                    } else {
                        emit("\t" . $initval);
                        last;
                    }
                }
            } else {
                emit $initval;
            }
        }
    }
    emit ";\n";
}

sub output_decls {
    my $i;

    DECL:
    for ($i = 0; $i <= $#symname; $i++) {
        foreach $parindex (@param) {
            next DECL if ($i == $parindex);
        }
        next if ($sympkg[$i] ne $main_pkgname or $symcommon[$i] ne "");
        output_decl $i;
    }

    # Output of COMMON blocks
    foreach $comname (keys %commons) {
        next unless ($commain{$comname});
        indent("type " . $comname . "_common is record\n");
        my $indexes = $commons{$comname};
        foreach $i (@$indexes) {
            output_decl $i;
        }
        dedent "end record;\n";
        dent($comname . " : " . $comname . "_common;\n\n");
    }
}

sub get_implicit_type {
    my $implic_index = ord(uc shift) - 65;
    if ($implicit_none or $implic_index < 0 or $implic_index > 25) {
        return $NONE;
    }
    $implic[$implic_index];
}

# initializations upon parsing a new SUBROUTINE

sub init_variables {
    $outbuf = "";
    $direct_output = 0;
    $indentlevel = 2;
    $n_symbols = 0;
    @symname = ();
    @symtype = ();
    @syminit = ();
    @symisconst = ();
    @symcommon = ();
    @symwritten = ();
    @symdim = ();
    @param = ();
    %commons = ();
    %commain = ();
    $subprogram_name = "";
    $functiontype = $NONE;
    $declarative_part = 1;
    $in_structure = 0;
    $in_blockdata = 0;
    $comment = "";
    $elsif = 0;
    $pending_endif = 0;
    $pending_label = 0;
    $arith_if_operator = "";
    $chardim = 0;
    $decldim = "";
}

# program startup initializations

sub init {
    $NONE       = -1;
    $CHAR       = 0;
    $LOGICAL1   = 1;
    $LOGICAL2   = 2;
    $LOGICAL    = 3;
    $INT1       = 4;
    $INT2       = 5;
    $INT        = 6;
    $REAL       = 7;
    $DOUBLE     = 8;
    $COMPLEX    = 9;
    $COMPLEX16  = 10;
    $STRING     = 11;
    @adatype =  ('Character',
                 'Logical1',
                 'Logical2',
                 'Logical',
                 'Integer1',
                 'Integer2',
                 'Integer',
                 'Real',
                 'Double',
                 'Complex',
                 'Complex16',
                 'Character_Star');
    $TO_BODY = 1;
    $TO_SPEC = 2;
    $eof = 0;
    $linebuf = "";
    $line_pending = 0;
    $first = 1;
    $seen_any_subprogram = 0;
    $opened_specfile = 0;
    $opened_bodyfile = 0;
    %functions = ();
    init_variables;
}

# Initialization for processing of IMPLICIT

sub init_implicit {      # must be called AFTER sub 'init'.
    @implic = (
        $REAL, $REAL, $REAL, $REAL, $REAL, $REAL, $REAL, $REAL,  # 'A'..'H'
        $INT,  $INT,  $INT,  $INT,  $INT,  $INT,                 # 'I'..'N'
        $REAL, $REAL, $REAL, $REAL, $REAL, $REAL, $REAL, $REAL,  # 'O'..'Z'
        $REAL, $REAL, $REAL, $REAL);
    $implicit_none = 0;
}

sub get_f77type_index {
    my $typestr = shift;
    my $stringlen = substr($typestr, 9);
    for ($typestr) {
        /^CHARACTER[1-9]/i and $type = $STRING + $stringlen, last;
        /^CHARACTER/i   and ($type = ($decldim eq "") ? $CHAR : $STRING), last;
        /^LOGICAL1/i    and $type = $LOGICAL1, last;
        /^LOGICAL2/i    and $type = $LOGICAL2, last;
        /^LOGICAL/i     and $type = $LOGICAL, last;
        /^INTEGER1/i    and $type = $INT1, last;
        /^INTEGER2/i    and $type = $INT2, last;
        /^INTEGER/i     and $type = $INT, last;
        /^DOUBLE/i      and $type = $DOUBLE, last;
        /^REAL8/i       and $type = $DOUBLE, last;
        /^REAL/i        and $type = $REAL, last;
        /^COMPLEX16/i   and $type = $COMPLEX16, last;
        /^COMPLEX/i     and $type = $COMPLEX, last;
        $type = $NONE, last;
    }
    $type;
}

sub basetype {
    my $type_index = shift;
    my $dimref = shift;
    unless ($dimref && @$dimref) {
        return($adatype[$type_index]);
    }
    my @dim = @$dimref;
    my $typename = $adatype[$type_index] . "_Array";
    my $dimensions = scalar(@dim);
    if ($dimensions > 1) {
       $typename .= '_' . $dimensions . 'D';
    }
    $typename;
}

sub targettype {
    my $type_index = shift;
    my $dimref = shift;
    my $typename;

    if ($type_index >= $STRING) {
        my $strlength = $type_index - $STRING;
        my $retval = 'Character';
        if ($strlength) {
            $retval .= $strlength;
        } else {
            $retval .= "_Star";
        }
        return $retval;
    }

    $typename = basetype($type_index, $dimref);

    unless ($dimref && @$dimref) {
        return($typename);
    }

    my @dim = @$dimref;
    if ($dim[0] eq '*') {
        if (scalar(@dim) > 1) {
            print "f2a restriction: declaration at $typename too complex\n";
        }
        return $typename;
    }
    $typename .= '(';
    my $i;
    for ($i = 0; $i <= $#dim; $i++) {
        my @bound = split /:/, $dim[$i];
        if (scalar(@bound) < 2) {
            $bound[1] = $bound[0];
            $bound[0] = 1;
        }
        foreach $b (qw/0 1/) {
            next if ($bound[$b] =~ /^[A-z]/);
            my $index = symlookup($bound[$b]);
            if ($index >= 0) {
                if ($sympkg[$index] ne $main_pkgname) {
                    $bound[$b] = $sympkg[$index] . '.' . $bound[$b];
                }
            }
        }
        $typename .= $bound[0] . " .. " . $bound[1];
        if ($i < $#dim) {
            $typename .= ',';
        }
    }
    $typename .= ')';
    $typename;
}

# ============================================================================

# sub constrained_type {
#     my $type_index = shift;
#     my $dimref = shift;
#     my $typename = basetype($type_index, $dimref);
# 
#     return($typename) unless ($dimref && @$dimref);
#     foreach $i (@dim) {
#         $typename .= '_' . $i;
#     }
#     return $typename;
# }

# various conversion specific utilities

sub check_begin {
    if ($declarative_part) {
        $declarative_part = 0;
        lessdent "begin\n";
    }
}

sub do_dim_initvals {
    my $dimref = shift;
    my $argref = shift;
    my $initstr_ref = shift;
    my @dims = @{$dimref};
    my @range = split(/:/, shift @dims);
    if (scalar(@range) < 2) {
        $range[1] = $range[0];
        $range[0] = 1;
    }
    my $len = $range[1] - $range[0] + 1;
    if (! scalar(@dims)) {
        $$initstr_ref .= '(';
        my $i;
        for ($i = 0; $i < $len; $i++) {
            while (scalar(@{$argref}) and $$argref[0] !~ /^[,\/]$/) {
                $$initstr_ref .= shift @{$argref};
            }
            shift @{$argref};
            if ($i < $len - 1) {
                $$initstr_ref .= ',';
            }
        }
        $$initstr_ref .= ')';
    } else {
        $$initstr_ref .= '(';
        while (--$len >= 0) {
            do_dim_initvals(\@dims, $argref, $initstr_ref);
            if ($len) {
                $$initstr_ref .= ',';
            }
        }
        $$initstr_ref .= ')';
    }
}

sub arith_if_comparison {
    # See if we can optimize an arithmetic IF.
    # Return one of $ltlab, $eqlab, $gtlab if we can optimize, else 0.
    # The returned label value is the one that must remain (i.e.
    # the one that is not optimized away.)
    # Further, the global variable $arith_if_operator is set to the
    # comparison operator string to use.
    my $ltlab = shift;
    my $eqlab = shift;
    my $gtlab = shift;
    my $following_lab = shift;

    $arith_if_operator = "";
    if ($ltlab == $eqlab) {
        if ($following_lab == $ltlab) {
            $arith_if_operator = ">";
            return $gtlab;
        } else {
            $arith_if_operator = "<=";
            return $ltlab;
        }
    }
    if ($ltlab == $gtlab) {
        if ($following_lab == $ltlab) {
            $arith_if_operator = "=";
            return $eqlab;
        } else {
            $arith_if_operator = "/=";
            return $ltlab;
        }
    }
    if ($eqlab == $gtlab) {
        if ($following_lab == $gtlab) {
            $arith_if_operator = "<";
            return $ltlab;
        } else {
            $arith_if_operator = ">=";
            return $gtlab;
        }
    }
    0;
}

